@@ -1,4 +1,5 @@
-{-# LANGUAGE TupleSections, ConstraintKinds #-}
+{-# LANGUAGE CPP, TupleSections, ConstraintKinds #-}
+{-# OPTIONS_GHC -fno-warn-duplicate-exports #-}
 
 -- | This module extends "Data.List" with extra functions of a similar nature.
 --   The package also exports the existing "Data.List" functions.
@@ -25,13 +26,13 @@ module Data.List.Extra(
     -- * List operations
     groupSort, groupSortOn, groupSortBy,
     nubOrd, nubOrdBy, nubOrdOn,
-    nubOn, groupOn,
+    nubOn, groupOn, groupOnKey,
     nubSort, nubSortBy, nubSortOn,
     maximumOn, minimumOn,
     sum', product',
     sumOn', productOn',
     disjoint, disjointOrd, disjointOrdBy, allSame, anySame,
-    repeatedly, firstJust,
+    repeatedly, repeatedlyNE, firstJust,
     concatUnzip, concatUnzip3,
     zipFrom, zipWithFrom, zipWithLongest,
     replace, merge, mergeBy,
@@ -48,11 +49,16 @@ import Numeric
 import Data.Functor
 import Data.Foldable
 import Prelude
+import Data.List.NonEmpty (NonEmpty ((:|)))
 
 
 -- | Apply some operation repeatedly, producing an element of output
 --   and the remainder of the list.
 --
+-- When the empty list is reached it is returned, so the operation
+-- is /never/ applied to the empty input.
+-- That fact is encoded in the type system with 'repeatedlyNE'
+--
 -- > \xs -> repeatedly (splitAt 3) xs  == chunksOf 3 xs
 -- > \xs -> repeatedly word1 (trim xs) == words xs
 -- > \xs -> repeatedly line1 xs == lines xs
@@ -61,6 +67,16 @@ repeatedly f [] = []
 repeatedly f as = b : repeatedly f as'
     where (b, as') = f as
 
+-- | Apply some operation repeatedly, producing an element of output
+--   and the remainder of the list.
+--
+-- Identical to 'repeatedly', but has a more precise type signature.
+repeatedlyNE :: (NonEmpty a -> (b, [a])) -> [a] -> [b]
+repeatedlyNE f [] = []
+repeatedlyNE f (a : as) = b : repeatedlyNE f as'
+    where (b, as') = f (a :| as)
+
+
 
 -- | Are two lists disjoint, with no elements in common.
 --
@@ -72,7 +88,7 @@ disjoint xs = null . intersect xs
 -- | /O((m+n) log m), m <= n/. Are two lists disjoint, with no elements in common.
 --
 -- @disjointOrd@ is more strict than `disjoint`. For example, @disjointOrd@ cannot
--- terminate if both lists are inifite, while `disjoint` can.
+-- terminate if both lists are infinite, while `disjoint` can.
 --
 -- > disjointOrd [1,2,3] [4,5] == True
 -- > disjointOrd [1,2,3] [4,1] == False
@@ -141,6 +157,7 @@ lastDef :: a -> [a] -> a
 lastDef d xs = foldl (\_ x -> x) d xs -- I know this looks weird, but apparently this is the fastest way to do this: https://hackage.haskell.org/package/base-4.12.0.0/docs/src/GHC.List.html#last
 {-# INLINE lastDef #-}
 
+#if __GLASGOW_HASKELL__ <= 906
 -- | A total variant of the list index function `(!!)`.
 --
 -- > [2,3,4] !? 1    == Just 3
@@ -154,6 +171,7 @@ xs !? n
                                    0 -> Just x
                                    _ -> r (k-1)) (const Nothing) xs n
 {-# INLINABLE (!?) #-}
+#endif
 
 -- | A composition of 'not' and 'null'.
 --
@@ -172,6 +190,7 @@ list :: b -> (a -> [a] -> b) -> [a] -> b
 list nil cons [] = nil
 list nil cons (x:xs) = cons x xs
 
+#if __GLASGOW_HASKELL__ <= 906
 -- | If the list is empty returns 'Nothing', otherwise returns the 'init' and the 'last'.
 --
 -- > unsnoc "test" == Just ("tes",'t')
@@ -182,6 +201,7 @@ unsnoc [] = Nothing
 unsnoc [x] = Just ([], x)
 unsnoc (x:xs) = Just (x:a, b)
     where Just (a,b) = unsnoc xs
+#endif
 
 -- | Append an element to the start of a list, an alias for '(:)'.
 --
@@ -271,7 +291,7 @@ zipWithFrom f a = zipWith f [a..]
 --
 -- > concatUnzip [("a","AB"),("bc","C")] == ("abc","ABC")
 concatUnzip :: [([a], [b])] -> ([a], [b])
-concatUnzip = (concat *** concat) . unzip
+concatUnzip = (concat *** concat) . Prelude.unzip
 
 -- | A merging of 'unzip3' and 'concat'.
 --
@@ -411,12 +431,26 @@ unescapeJSON [] = []
 
 
 -- | A version of 'group' where the equality is done on some extracted value.
-groupOn :: Eq b => (a -> b) -> [a] -> [[a]]
+--
+-- > groupOn abs [1,-1,2] == [[1,-1], [2]]
+groupOn :: Eq k => (a -> k) -> [a] -> [[a]]
 groupOn f = groupBy ((==) `on2` f)
     -- redefine on so we avoid duplicate computation for most values.
     where (.*.) `on2` f = \x -> let fx = f x in \y -> fx .*. f y
 
 
+-- | A version of 'groupOn' which pairs each group with its "key" - the
+--   extracted value used for equality testing.
+--
+-- > groupOnKey abs [1,-1,2] == [(1, [1,-1]), (2, [2])]
+groupOnKey :: Eq k => (a -> k) -> [a] -> [(k, [a])]
+groupOnKey _ []     = []
+groupOnKey f (x:xs) = (fx, x:yes) : groupOnKey f no
+    where
+        fx = f x
+        (yes, no) = span (\y -> fx == f y) xs
+
+
 -- | /DEPRECATED/ Use 'nubOrdOn', since this function is _O(n^2)_.
 --
 --   A version of 'nub' where the equality is done on some extracted value.
@@ -462,7 +496,8 @@ minimumOn f (x:xs) = g x (f x) xs
 -- > \xs -> map fst (groupSort xs) == sort (nub (map fst xs))
 -- > \xs -> concatMap snd (groupSort xs) == map snd (sortOn fst xs)
 groupSort :: Ord k => [(k, v)] -> [(k, [v])]
-groupSort = map (\x -> (fst $ head x, map snd x)) . groupOn fst . sortOn fst
+groupSort = map (\x -> (fst $ headErr x, map snd x)) . groupOn fst . sortOn fst
+    where headErr (x:_) = x
 
 -- | A combination of 'group' and 'sort', using a part of the value to compare on.
 --
@@ -520,15 +555,17 @@ mergeBy f (x:xs) (y:ys)
     | otherwise = y : mergeBy f (x:xs) ys
 
 
--- | Replace a subsequence everywhere it occurs. The first argument must
---   not be the empty list.
+-- | Replace a subsequence everywhere it occurs.
 --
 -- > replace "el" "_" "Hello Bella" == "H_lo B_la"
 -- > replace "el" "e" "Hello"       == "Helo"
--- > replace "" "e" "Hello"         == undefined
--- > \xs ys -> not (null xs) ==> replace xs xs ys == ys
-replace :: (Partial, Eq a) => [a] -> [a] -> [a] -> [a]
-replace [] _ _ = error "Extra.replace, first argument cannot be empty"
+-- > replace "" "x" "Hello"         == "xHxexlxlxox"
+-- > replace "" "x" ""              == "x"
+-- > \xs ys -> replace xs xs ys == ys
+replace :: Eq a => [a] -> [a] -> [a] -> [a]
+replace [] to xs = go xs
+    where go [] = to
+          go (x:xs) = to ++ x : go xs
 replace from to xs | Just xs <- stripPrefix from xs = to ++ replace from to xs
 replace from to (x:xs) = x : replace from to xs
 replace from to [] = []
@@ -865,14 +902,41 @@ zipWithLongest f (x:xs) (y:ys) = f (Just x) (Just y) : zipWithLongest f xs ys
 zipWithLongest f [] ys = map (f Nothing . Just) ys
 zipWithLongest f xs [] = map ((`f` Nothing) . Just) xs
 
--- | Lazily compare the length of a 'Foldable' with a number.
+#if __GLASGOW_HASKELL__ <= 910
+-- | Use 'compareLength' @xs@ @n@ as a safer and faster alternative
+-- to 'compare' ('length' @xs@) @n@. Similarly, it's better
+-- to write @compareLength xs 10 == LT@ instead of @length xs < 10@.
 --
--- > compareLength [1,2,3] 1 == GT
--- > compareLength [1,2] 2 == EQ
--- > \(xs :: [Int]) n -> compareLength xs n == compare (length xs) n
--- > compareLength (1:2:3:undefined) 2 == GT
-compareLength :: (Ord b, Num b, Foldable f) => f a -> b -> Ordering
-compareLength = foldr (\_ acc n -> if n > 0 then acc (n - 1) else GT) (compare 0)
+-- While 'length' would force and traverse
+-- the entire spine of @xs@ (which could even diverge if @xs@ is infinite),
+-- 'compareLength' traverses at most @n@ elements to determine its result.
+--
+-- >>> compareLength [] 0
+-- EQ
+-- >>> compareLength [] 1
+-- LT
+-- >>> compareLength ['a'] 1
+-- EQ
+-- >>> compareLength ['a', 'b'] 1
+-- GT
+-- >>> compareLength [0..] 100
+-- GT
+-- >>> compareLength undefined (-1)
+-- GT
+-- >>> compareLength ('a' : undefined) 0
+-- GT
+--
+-- @since 4.21.0.0
+--
+compareLength :: [a] -> Int -> Ordering
+compareLength xs n
+  | n < 0 = GT
+  | otherwise = foldr
+    (\_ f m -> if m > 0 then f (m - 1) else GT)
+    (\m -> if m > 0 then LT else EQ)
+    xs
+    n
+#endif
 
 -- | Lazily compare the length of two 'Foldable's.
 -- > comparingLength [1,2,3] [False] == GT
